@@ -1,6 +1,6 @@
 ;; support.scm -- Various support facilities, used by herd and shepherd.
 ;; Copyright (C) 2014 A.Sassmannshausen <alex.sassmannshausen@gmail.com>
-;; Copyright (C) 2013, 2014, 2016, 2018, 2020 Ludovic Courtès <ludo@gnu.org>
+;; Copyright (C) 2013-2014, 2016, 2018, 2020, 2022 Ludovic Courtès <ludo@gnu.org>
 ;; Copyright (C) 2002, 2003 Wolfgang Jährling <wolfgang@pro-linux.de>
 ;; Copyright (C) 2016 Mathieu Lirzin <mthl@gnu.org>
 ;; Copyright (C) 2018 Danny Milosavljevic <dannym@scratchpost.org>
@@ -23,6 +23,7 @@
 (define-module (shepherd support)
   #:use-module (shepherd config)
   #:use-module (ice-9 match)
+  #:use-module (ice-9 format)
   #:export (buffering-mode
 
             caught-error
@@ -44,6 +45,7 @@
             program-name
             report-error
             display-line
+            non-blocking-port
 
             user-homedir
             user-default-log-file
@@ -60,24 +62,19 @@
             persistency
             persistency-state-file
 
-            verify-dir))
+            %user-log-dir
+            %user-config-dir
+            %user-runtime-dir
 
-(define-syntax-rule (if-2.0 subsequent alternate)
-  "Expand to SUBSEQUENT when using Guile 2.0, and to ALTERNATE otherwise."
-  (cond-expand
-    ((and guile-2 (not guile-2.2)) subsequent)
-    (else alternate)))
+            verify-dir))
 
 (define-syntax buffering-mode
   (syntax-rules (line block none)
     "Return the appropriate buffering mode depending on whether we're on Guile
 2.0 or later."
-    ((_ line)
-     (if-2.0 _IOLBF 'line))
-    ((_ block)
-     (if-2.0 _IOFBF 'block))
-    ((_ none)
-     (if-2.0 _IONBF 'none))))
+    ((_ line) 'line)
+    ((_ block) 'block)
+    ((_ none) 'none)))
 
 ;; Report the caught error.
 ;; FIXME: Needs some more work.
@@ -205,17 +202,16 @@ output port, and PROC's result is returned."
   (unless (getenv "GUILE_WARN_DEPRECATED")
     (debug-disable 'warn-deprecated))
 
-  ;; In Guile 2.2+, the locale is installed by default.
-  (if-2.0 (false-if-exception (setlocale LC_ALL ""))
-          #t)
-
   (bindtextdomain %gettext-domain %localedir)
   (textdomain %gettext-domain)
   (setvbuf (current-output-port) (buffering-mode line))
   (setvbuf (current-error-port) (buffering-mode line)))
 
 ;; Localized version of STR.
-(define l10n gettext)
+(define-syntax l10n
+  (syntax-rules ()
+    ((_ str) (gettext str))
+    ((_ singular plural n) (ngettext singular plural n))))
 
 ;; Display some text and a newline.
 (define-syntax-rule (local-output format-string args ...)
@@ -226,7 +222,7 @@ output port, and PROC's result is returned."
 (define* (display-version #:optional (program-name (program-name)))
   (local-output "~a (~a) ~a" program-name package-name Version)
   ;; TRANSLATORS: '~a' is a placeholder for the copyright year.
-  (local-output (l10n "Copyright (C) ~a the Shepherd authors") 2020)
+  (local-output (l10n "Copyright (C) ~a the Shepherd authors") 2022)
   (local-output (l10n "\
 License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>
 This is free software: you are free to change and redistribute it.
@@ -252,6 +248,12 @@ There is NO WARRANTY, to the extent permitted by law.")))
   (display message port)
   (newline port))
 
+(define (non-blocking-port port)
+  "Return PORT after putting it in non-blocking mode."
+  (let ((flags (fcntl port F_GETFL)))
+    (fcntl port F_SETFL (logior O_NONBLOCK flags))
+    port))
+
 
 
 ;; Home directory of the user.
@@ -271,6 +273,12 @@ There is NO WARRANTY, to the extent permitted by law.")))
                      (string-append user-homedir "/.config"))
                  "/shepherd"))
 
+(define %user-log-dir
+  ;; Default log directory if shepherd is run as a normal user.
+  (string-append (or (getenv "XDG_STATE_HOME")
+                     (string-append user-homedir "/.local/state"))
+                 "/shepherd"))
+
 (define %user-runtime-dir
   ;; Default runtime directory if shepherd is run as a normal user.
   (string-append (or (getenv "XDG_RUNTIME_DIR")
@@ -282,28 +290,32 @@ There is NO WARRANTY, to the extent permitted by law.")))
 TARGET should be a string representing a filepath + name."
   (with-output-to-file target
     (lambda ()
-      (display (string-append
-                ";; init.scm -- default shepherd configuration file.
-
+      (display
+       (string-append
+        ;; TRANSLATORS: Please keep double semicolons at the beginning of each
+        ;; line: they introduce comments in the Scheme language.  Also, keep
+        ;; "shepherd" untranslated.  Thank you!  :-)
+        (l10n ";; init.scm -- default shepherd configuration file.\n")
+        "\n"
+        (l10n "\
 ;; Services known to shepherd:
 ;; Add new services (defined using 'make <service>') to shepherd here by
-;; providing them as arguments to 'register-services'.
-""(register-services)
-
-;; Send shepherd into the background
-""(action 'shepherd 'daemonize)
-
+;; providing them as arguments to 'register-services'.\n")
+        "(register-services)\n\n"
+        (l10n "\
+;; Send shepherd into the background\n")
+        "(action 'shepherd 'daemonize)\n\n"
+        (l10n "\
 ;; Services to start when shepherd starts:
 ;; Add the name of each service that should be started to the list
-;; below passed to 'for-each'.
-""(for-each start '())
-")))))
+;; below passed to 'for-each'.\n")
+        "(for-each start '())\n")))))
 
 ;; Logging.
 (define (user-default-log-file)
   "Return the file name of the user's default log file."
-  (mkdir-p %user-config-dir #o700)
-  (string-append %user-config-dir "/shepherd.log"))
+  (mkdir-p %user-log-dir #o700)
+  (string-append %user-log-dir "/shepherd.log"))
 
 (define default-logfile-date-format
   ;; 'strftime' format string to prefix each entry in the log.
